#include "mcsClustering.h"


////////////////////////////////////////////////////////////////////////////////////
// class ClusterMolecule
//    derived class from Molecule with all stuff for clustering
//    which is molecule specific
////////////////////////////////////////////////////////////////////////////////////
const ClusterMolecule::PairType ClusterMolecule::convertStructure(const char& myStr)
{
   switch(myStr)
   {
      case '.':
         return UNPAIRED;
         break;
      case '(':
         return LEFTPAIRED;
         break;
      case ')':
         return RIGHTPAIRED;
         break;
      default:
         return UNDEFPAIR;  
   }
}

void ClusterMolecule::initSequence()
{
   // init mol.sequence with 0,1..n,n+1 nucleotides
   // 0,n+1 are additional nucledtides for algo with undef structure & bond
   // 0,n+1 is closing bond for all unbound nucs at level "0" (dangling ends + others) 
   sequence.clear();
   NucleotidePTR myNuc;
   for (unsigned int i=0;i<=getSequence().length()+1;i++)
   {  
      myNuc = NucleotidePTR(new Nucleotide);
      myNuc->pairType   		= convertStructure(getStructure(i));
      myNuc->structureType		= UNDEFSTRUCTURE;
      myNuc->position    		= i;
      myNuc->connectedTo 		= this->getBasePartner(i);
      myNuc->lastClosingBond 	= make_pair(0,this->getSequence().length()+1);
      
      sequence.push_back(myNuc);
   }
   // calculate all closing bonds from secondary-structure elements
   // (hairpins,internal-loops,multiloops, stems)
   // only for strand[1]..[n]
   NucleotidePTR prevNuc,actNuc;
   std::stack<pair<unsigned int,unsigned int> > bondStack;
   bondStack.push(make_pair(0,sequence.size()-1));
   for (unsigned int i=1;i<sequence.size()-1;i++)
   {
      prevNuc = sequence[i-1];
      actNuc  = sequence[i];
      
      switch (actNuc->pairType)
      {
         case LEFTPAIRED:
            if ( prevNuc->pairType != LEFTPAIRED )
            {
               actNuc->lastClosingBond = prevNuc->lastClosingBond;
               bondStack.push(std::make_pair(actNuc->position,actNuc->connectedTo));
               actNuc->structureType = STEM_OUTSIDE_MIN;
            }
            if ( prevNuc->pairType == LEFTPAIRED )
               if (abs(prevNuc->connectedTo - actNuc->connectedTo) != 1)
               {
                  if (bondStack.top().first != prevNuc->position)    //case if prevNuc = STEM_ENDBOND
                     prevNuc->lastClosingBond = intPair(bondStack.top().first,bondStack.top().second); 
                  bondStack.push(std::make_pair(prevNuc->position,prevNuc->connectedTo));
                  prevNuc->structureType = STEM_INSIDE_END;
                 
                  actNuc->lastClosingBond = intPair(bondStack.top().first,bondStack.top().second);
                  bondStack.push(std::make_pair(actNuc->position,actNuc->connectedTo));
                  actNuc->structureType = STEM_OUTSIDE_MIN;
               }
               else 
               {  
                  actNuc->structureType = STACKING;
                  actNuc->lastClosingBond = intPair(bondStack.top().first,bondStack.top().second);
               }
            break;
         case UNPAIRED:
            if ( prevNuc->pairType == LEFTPAIRED )
            {
               bondStack.push(std::make_pair(prevNuc->position,prevNuc->connectedTo));
               prevNuc->structureType = STEM_INSIDE_END;
               actNuc->lastClosingBond = intPair(bondStack.top().first,bondStack.top().second);
            }
            if ( prevNuc->pairType == RIGHTPAIRED )
            {
               bondStack.pop();               
               prevNuc->structureType = STEM_OUTSIDE_MAX;
               actNuc->lastClosingBond = prevNuc->lastClosingBond;
            }
            if ( prevNuc->pairType == UNPAIRED )
            {
               actNuc->lastClosingBond = prevNuc->lastClosingBond;
            }
            actNuc->structureType = LOOP;
            break;
         case RIGHTPAIRED:
            if (abs(prevNuc->connectedTo - actNuc->connectedTo) != 1)
            {
               bondStack.pop();
               actNuc->structureType = STEM_INSIDE_END;
            }
            sequence[i]->lastClosingBond = sequence[sequence[i]->connectedTo]->lastClosingBond;
            break;
         case UNDEFPAIR: 
            cerr << "Error makeStrand!" << endl;
      }
   }
   /*for (SequenceITER i = sequence.begin();i!=sequence.end();++i)
   {
      NucleotidePTR myNuc;
      myNuc = *i;
      cout << "\t-------NUCLEOTIDE-------------------------------------"<<endl;
   	cout << "\t\t\tPosition : " << myNuc->position << endl;
   	cout << "\t\t\tStructure: " << myNuc->structureType << endl;
   	cout << "\t\t\tConnected: " << myNuc->connectedTo << endl;
   	cout << "\t\t\tClosingB : " << myNuc->lastClosingBond.first <<","<< myNuc->lastClosingBond.second<< endl;
   	cout << "\t\t\tpairType : " << myNuc->pairType<<endl;
   	cout << "\t------------------------------------------------------\n";
   }*/
}

////////////////////////////////////////////////////////////////////////////////////
// class mcsClustering
////////////////////////////////////////////////////////////////////////////////////

// main call from mcsEnsemble to start everything
void mcsClustering::calculateClusters()
{
   clusterCounter = 0;
   
   if (DISTANCE_METHODE != DIST_EQUAL)
	   DISTANCE_EQUAL_THRESHOLD = DISTANCE_THRESHOLD * 2;
   
   preProcessing();
   clustering();
   
   collectClusters();
}

// init the clusterSequence data structure - pattern ordering step
// 	1.	calculate structure elements and closing bonds from both structures (mol.initSequence)
//		2. init ClusterSequence with ClusterNucleotides
//			2a. init beforeNuc/afterNuc/connectedToNuc pointers
//		3. insert all pattern keys - addPatternKey
void mcsClustering::preProcessing()
{
   // 1. create for each molecule the sequence of special nucleotides for clustering 
   mol1.initSequence();
   mol2.initSequence();
   
   // 2. init the main clusterSequence with the nucleotides from molecule 1  
   clusterSequence.clear();
   ClusterNucleotidePTR myClusterNucleotideInit;

   for (ClusterMolecule::SequenceITER myMol1SeqNuc = mol1.sequence.begin();myMol1SeqNuc != mol1.sequence.end();++myMol1SeqNuc)
   {
   	myClusterNucleotideInit 									= ClusterNucleotidePTR(new ClusterNucleotide);
   	myClusterNucleotideInit->nucleotide 					= *myMol1SeqNuc;
   	myClusterNucleotideInit->clusterNucleotideType 		= CLUSTER_TO_BEFORE_BASE;
   	myClusterNucleotideInit->beforeNuc						= myClusterNucleotideInit;;
   	myClusterNucleotideInit->afterNuc						= myClusterNucleotideInit;;
   	myClusterNucleotideInit->connectedToNuc				= myClusterNucleotideInit;
   	myClusterNucleotideInit->clusterList.clear();
   	myClusterNucleotideInit->patternKeyList.clear();
   	
   	clusterSequence.push_back(myClusterNucleotideInit);
   }
   
   // 2a. init beforeNuc/afterNuc/connectedToNuc pointers
   for (ClusterSequenceCITER myClusNuc = clusterSequence.begin(); myClusNuc != clusterSequence.end();++myClusNuc)
   {   	
	if ((*myClusNuc)->nucleotide->position != 0)
   		clusterSequence[(*myClusNuc)->nucleotide->position]->beforeNuc = clusterSequence[(*myClusNuc)->nucleotide->position-1];
   	if ((*myClusNuc)->nucleotide->position < (mol1.sequence.size()-1))
   		clusterSequence[(*myClusNuc)->nucleotide->position]->afterNuc = clusterSequence[(*myClusNuc)->nucleotide->position+1];
   	if ((*myClusNuc)->nucleotide->connectedTo != -1)
   		clusterSequence[(*myClusNuc)->nucleotide->position]->connectedToNuc = clusterSequence[(*myClusNuc)->nucleotide->connectedTo];
   }

   // 3. insert pattern keys in the clusterSequence from all patterns -> ordering for the first/main structure
   // 	- calculate all necessary values (calculatePatternBoundaries) from pattern 
   // 	and create PatternKey datastructure "myKey" for it 
   // 	- then insert myKey in clusterSequence[max-outside-bound].patternKeyList
   // 	- key-pos is always maximum position of the pattern
   for (PatternPairMap::patListCITER myPair = patterns.getList().begin(); myPair != patterns.getList().end(); ++myPair)
   {
    PatternKeyPTR myKey = PatternKeyPTR(new PatternKey());
   	
   	calculatePatternBoundaries(*myPair, myKey);         
 		
   	myKey->position      		= myKey->minmax.second;			// at maximum 
   	myKey->patternId         	= (*myPair)->getId();
    myKey->patternSize       	= (*myPair)->getSize();                                 
    //myKey->lastClosingBond     = mol1.sequence[myKey->minmax.second]->lastClosingBond;
    //myKey->lastClosingBondREF  = mol2.sequence[myKey->minmaxREF.second]->lastClosingBond;
    
    switch (myKey->patternType)
    {
    case LOOPPATTERN:
    	myKey->lastClosingBond     = mol1.sequence[myKey->minmax.second]->lastClosingBond;
    	myKey->lastClosingBondREF  = mol2.sequence[myKey->minmaxREF.second]->lastClosingBond;
    	break;
    default:
    	myKey->lastClosingBond     = myKey->minmax;
    	myKey->lastClosingBondREF  = myKey->minmaxREF;
    	break;
    }
   
    // insert Key in clusterSequence
    clusterSequence[myKey->position]->patternKeyList.push_back(myKey);
   }

/*   //priting clusterSequence info
   for (ClusterSequenceCITER myClusNuc = clusterSequence.begin(); myClusNuc != clusterSequence.end();++myClusNuc)
   {
   	cout << "------------------------------->" << endl;
   	cout << "Position : " << (*myClusNuc)->nucleotide->position <<endl;
   	cout << "ConnectTo: " << (*myClusNuc)->nucleotide->connectedTo << endl;
   	cout << "BeforeNuc: " << (*myClusNuc)->beforeNuc->nucleotide->position << endl;
   	cout << "AfterNuc : " << (*myClusNuc)->afterNuc->nucleotide->position << endl;
   	cout << "ToNuc    : " << (*myClusNuc)->connectedToNuc->nucleotide->position << endl;
   	cout << "PatKeys  : " << (*myClusNuc)->patternKeyList.size() << endl;
   }*/
}


// determines boundaries and type/shape of a pattern
//   LOOPPATTERN    : part of a loop or hairpin
//   STEMPATTERN    : all patterns with at least 1 bond and all subkeys are NOT unbound
//   SUBLOOPPATTERN : all STEMPATTERNs with an unbound subkey 
//void mcsClustering::calculatePatternBoundaries(const	boost::shared_ptr<PatternPair>& myPair,
//                                                      PatternKeyPTR myKey)
void mcsClustering::calculatePatternBoundaries(const	PatternPair* myPair,
                                                      PatternKeyPTR myKey)


{
   intVec preBounds,preBoundsREF;
   bool withBond = false;
   const intVec& myPatStructure1 = myPair->getFirstPat().getPat();
   const intVec& myPatStructure2 = myPair->getSecPat().getPat();

   //determines the maximum extetend shape of the pattern (add all missing basepair partners)
   // REF is for the second structure
   for (unsigned int i=0;i<myPatStructure1.size();++i)
   {
      if (mol1.getStructure(myPatStructure1[i]) == '.') 
      {
         preBounds.push_back(myPatStructure1[i]);
         preBoundsREF.push_back(myPatStructure2[i]);
      }
         else
         {
            preBounds.push_back(myPatStructure1[i]);
            preBounds.push_back(mol1.getBasePartner(myPatStructure1[i]));        
            preBoundsREF.push_back(myPatStructure2[i]);
            preBoundsREF.push_back(mol2.getBasePartner(myPatStructure2[i]));
            withBond = true;
         }         
   }
   sort(preBounds.begin(),preBounds.end());
   sort(preBoundsREF.begin(),preBoundsREF.end());
   
   myKey->allBounds.clear();
   myKey->allBoundsREF.clear();
   myKey->minBounds.clear();
   myKey->maxBounds.clear();
   myKey->minBoundsREF.clear();
   myKey->maxBoundsREF.clear();

   // determine all boundaries
   // between two pattern positions have to be space for another pattern
   for (unsigned int k=1;k < (preBounds.size());++k)
   {
      if ((preBounds[k]-minPatternSize > preBounds[k-1])  
        &&(preBoundsREF[k]-minPatternSize > preBoundsREF[k-1])) 
      {             
         myKey->allBounds.push_back(preBounds[k-1]);
         myKey->allBounds.push_back(preBounds[k]);
         myKey->allBoundsREF.push_back(preBoundsREF[k-1]);
         myKey->allBoundsREF.push_back(preBoundsREF[k]);
         myKey->maxBounds.push_back(preBounds[k-1]);
         myKey->minBounds.push_back(preBounds[k]);
         myKey->maxBoundsREF.push_back(preBoundsREF[k-1]);
         myKey->minBoundsREF.push_back(preBoundsREF[k]);
      }
   }
   // insert global min/max of the pattern
   myKey->allBounds.insert(myKey->allBounds.begin(),preBounds.front());
   myKey->allBounds.push_back(preBounds.back());
   myKey->allBoundsREF.insert(myKey->allBoundsREF.begin(),preBoundsREF.front());
   myKey->allBoundsREF.push_back(preBoundsREF.back());
   
   // is it a loop-pattern? then fill min/max Bounds with pattern min/max
   if ((myKey->allBounds.size() == 2))//||((myKey->allBounds.size() == 2)&&(mol1.getStructure(myKey->allBounds.front()-1)!='(')))
      {
         myKey->minBounds.push_back(myKey->allBounds.front());
         myKey->maxBounds.push_back(myKey->allBounds.back());
         myKey->minBoundsREF.push_back(myKey->allBoundsREF.front());
         myKey->maxBoundsREF.push_back(myKey->allBoundsREF.back());
      }
   
 /*  if ((myKey->allBounds.size() == 2)&&(mol1.getStructure(myKey->allBounds.front()-1)!='('))
         {
            myKey->minBounds.insert(myKey->minBounds.begin(),myKey->allBounds.front());
            myKey->minBoundsREF.insert(myKey->minBoundsREF.begin(),myKey->allBoundsREF.front());
         }*/
   // min/max pair, so far this ist first/last element of allBounds
   myKey->minmax    = make_pair(myKey->allBounds.front(),myKey->allBounds.back());
   myKey->minmaxREF = make_pair(myKey->allBoundsREF.front(),myKey->allBoundsREF.back());
   
   
   //determines Pattern Type according to real shape for all patterns
   if (myKey->allBounds.size() > 2)
   { 
      myKey->patternType = STEMPATTERN;
      for (intVecCITER myBound  = myKey->allBounds.begin()+1;myBound != myKey->allBounds.end()-1;++myBound)
      {
         if (mol1.getStructure(*myBound) == '.')
            myKey->patternType = STEMLOOPPATTERN;
      }
      if ((myKey->allBounds.size()>4)&&
          (mol1.getStructure(myKey->allBounds[myKey->allBounds.size()-2])!='.')&&(mol1.getStructure(myKey->allBounds[1])!='.')&&
          (mol1.getBasePartner(myKey->allBounds[myKey->allBounds.size()-2])<=mol1.getBasePartner(myKey->allBounds[1])))
         myKey->patternType = STEMLOOPPATTERN;
   }
   else
   	if ((myKey->allBounds.size() == 2) && (!withBond))
   		myKey->patternType = LOOPPATTERN;
   	else
   		myKey->patternType = HAIRPINPATTERN;
   
  /* if ((myKey->boundaries.size()>4)||(myKey->patternType == STEMLOOPPATTERN))
   {
   cout << "------> Pattern Type: "<<myKey->patternType<<endl;
   cout << "\tpreBound1: "<< biu::Util_String::intvec2str(preBounds,":")<<endl;
   cout << "\tpreBound2: "<< biu::Util_String::intvec2str(preBoundsREF,":")<<endl;
   cout << "\tPattern1: "<< biu::Util_String::intvec2str(myPatStructure1,":")<<endl;
   cout << "\tminBound: "<< biu::Util_String::intvec2str(myKey->minBounds,":")<<endl;
   cout << "\tmaxBound: "<< biu::Util_String::intvec2str(myKey->maxBounds,":")<<endl;
   cout << "\tmin max : "<< myKey->minmax.first<<","<<myKey->minmax.second<<endl;
   cout << "\tBounds  : "<< biu::Util_String::intvec2str(myKey->allBounds,":")<<endl;
   cout << "\tPattern2: "<< biu::Util_String::intvec2str(myPatStructure2,":")<<endl;
   cout << "\tBounds2 : "<< biu::Util_String::intvec2str(myKey->allBoundsREF,":")<<endl;
   cout << "\tmin max : "<< myKey->minmaxREF.first<<","<<myKey->minmaxREF.second<<endl;
   cout << endl;
   cout<<endl;
   }*/
}


// main algorithm for clustering
//
// main iteration is over clusterStrand 
// clusterStrand is a vector of BaseClusterListS(baseType,nucleotide,clusterList)
// patterns keys are in mol1.clusterKeyStrand
//
// for each clusterStrand[i]
//		(1)	new patterns at pos i are clustered with method CLUSTER_METHODE
//			(1a)	if i.baseType is "CLUSTER_STEM_TO_BEFORE_BASE",
//					stem is clustered to base before stem start (STEM_START_LEFTBOND-1, i.connectedTo-1)			
//		(2)	clusterList i is updated from i-1 with DISTANCE_METHODE
//			(2a)	if i is STEMSTART_LEFTBOND, the STEMSTART_RIGHTBOND is marked 
//					as "CLUSTER_STEM_TO_BEFORE_BASE", clusterlist[i] is empty
//			(2b)	clusterList is also updated from  STEM_START_LEFTBOND-1 with DISTANCE_METHODE
//					(clusters before stem could be in treshold!!!)
void mcsClustering::clustering()
{
   ClusterSequenceITER pos,end;
   ClusterListTYPE   actClusterList;
   
   // set pos to clusterStrand[1]
   pos = ++clusterSequence.begin();
   
   // set end to clusterStrand[n]
   end = clusterSequence.end()--;
   end = end-1;
   
   for(;pos != end;++pos)
   {
      switch ((*pos)->clusterNucleotideType)
      {
        /* case CLUSTER_TO_LOOP_BASE:
               // make disjunct cluster lists for left/right part of loop and cluster keymap pos to these 2 lists
               // make a cluster list with all clusters within treshold for left and rigt part of loop
               // save this list as one list at pos, distance update is also made
               // no extra call to updateClusterList   
               // makedisjunctLists(pos_before);
            break;*/
         case CLUSTER_TO_BEFORE_BASE: 
         		// normal sequence base, no stem end
         		// cluster all new patterns with outside max bound at pos (clustersequence->patternKeyList)
            	// to possible clusters before/inside
               clusterActPatterns(*pos);
				break;         
         case CLUSTER_STEM_TO_BEFORE_BASE:
         		// pos is last base of stem (STEM_OUTSIDE_MAX)
         		// 1. clustering of all new patterns ending at pos within stem
         		// 2. clustering stem to possible clusters before stem 
         		clusterActPatterns(*pos);
         		clusterToCluster(*pos);             
            break;
         default:break;
      }	//end switch
      
      // update the actual clusterlist with CLUSTER_METHODE
      // 1. update from pos before
      // 2. if pos = STEM_OUTSIDE_MAX, then update also from pos before stem
      updateClusterList(*pos);
      
		// only info
      if ((*pos)->nucleotide->position%10 == 0)	cout << (*pos)->nucleotide->position<<endl;
   }	// end main for loop
   
   cout << "-----> end"<<endl;
   cout << "ClusterMethode:" << CLUSTER_METHODE<<endl;
   cout << "DistMethode   :" << DISTANCE_METHODE<<endl;
   cout << "Clusters found: " << clusterCounter<<endl; 
}

// update the cluster list for each nucleotide after clustering is done
// check all clusters if they are in DISTANCE_TRESHOLD for the new nucleotide
// each nucleotide: update from nuc before with dist+1
// stem end(stem_outside_max): update from pos before stem with DISTANCE_METHODE
void mcsClustering::updateClusterList(const ClusterNucleotidePTR& toPos)
{
   int distanceChange;
   
   // standard update for clusterList from one Nucleotide before
   // !(*myCluster)->clustered is for CLUSTER_METHOD = CLUSTER_MAX_2, 
   // clustered = FALSE, always for CLUSTER_MAX_1, clustered = TRUE, only possible for CLUSTER_MAX_2, these clusters are not updated to toPos
   
   switch (toPos->nucleotide->structureType)
   {
   	case ClusterMolecule::STEM_OUTSIDE_MIN:
   		toPos->clusterList.clear();
       	toPos->connectedToNuc->clusterNucleotideType = CLUSTER_STEM_TO_BEFORE_BASE;
   		break;
   	default:
   		distanceChange = 1;
   		for (ClusterListITER myCluster = toPos->beforeNuc->clusterList.begin();myCluster != toPos->beforeNuc->clusterList.end();++myCluster)
   		{
      		if (((*myCluster)->actDistance+distanceChange <= DISTANCE_THRESHOLD) && (!(*myCluster)->clustered))
      		{
      			(*myCluster)->actDistance += distanceChange;
         		toPos->clusterList.push_back(*myCluster);
      		}
		}
   		break;
   }
   
   // update after a stem from the nucleotide before the stem
   switch (toPos->clusterNucleotideType)
   {
   	case CLUSTER_STEM_TO_BEFORE_BASE:
   		switch (DISTANCE_METHODE)
   		{
   			case DIST_SEQUENCE:case DIST_EQUAL:
   				distanceChange = abs((int(toPos->nucleotide->position)) - (int(toPos->connectedToNuc->beforeNuc->nucleotide->position)));
   				break;
   			case DIST_SHORTESTPATH:
   				distanceChange = 2;
   				break;
   		}
   		for (ClusterListCITER myCluster = toPos->connectedToNuc->beforeNuc->clusterList.begin();myCluster != toPos->connectedToNuc->beforeNuc->clusterList.end();++myCluster)
      	{
         	if (((*myCluster)->actDistance+distanceChange <= DISTANCE_THRESHOLD)&&(!(*myCluster)->clustered))
         	{
            	(*myCluster)->actDistance += distanceChange;
            	toPos->clusterList.push_back(*myCluster);
         	}
      	}
      	break;
      default:
      	break;
   }
   
  /* // mark right base of next closing bond for later processing
      //fehler: from or to pos??????   
   if ((!toPos->clusterList.empty()) && 
       ((toPos->nucleotide->bond == ClusterMolecule::STEMSTART_RIGHTBOND)||
        (toPos->nucleotide->bond == ClusterMolecule::UNBOUND)))
      clusterStrand[toPos->nucleotide->lastClosingBond.second].baseType = CLUSTER_TO_LOOP_BASE;  */
}

void mcsClustering::clusterActPatterns(const ClusterNucleotidePTR& myPos)
{
   ClusterListTYPE 	newClusterList;
   ClusterInfoPTR 	myPCluster,myExtCluster;
   
   // make clusters from all patterns with maximum at myPos
   for(PatternKeyListCITER actPattern = myPos->patternKeyList.begin();actPattern != myPos->patternKeyList.end();++actPattern)
   {
      ++clusterCounter;
      myPCluster                    = ClusterInfoPTR(new ClusterInfoTYPE);
      myExtCluster                  = ClusterInfoPTR(new ClusterInfoTYPE);
      
      myPCluster->clusterId         = "cl_"+ biu::Util_String::int2str(clusterCounter);
      myPCluster->clusterPatternList.push_back(*actPattern);
      myPCluster->newPattern        = *actPattern;
      myPCluster->lastClusterList.clear();
      myPCluster->clusterMinMax     = (*actPattern)->minmax;
      myPCluster->clusterSize       = (*actPattern)->patternSize;
      myPCluster->clusterMinKey     = (*actPattern);
      myPCluster->clusterMaxKey     = (*actPattern);
      myPCluster->actDistance       = 0;
      myPCluster->weight            = 0;
      myPCluster->visited           = false;
      myPCluster->clustered			= false;
      
      // copy content
      *myExtCluster = *myPCluster;
      
      // add Pattern as Cluster to newClusterList
      // Hairpins are the most inner structures, no pattern could be before/inside a hairpin -> no clustering needed
      if ((*actPattern)->patternType != HAIRPINPATTERN) 
      	newClusterList.push_back(myExtCluster);
      // add pattern to clusterSequence as Cluster without clustering
      (*myPos).clusterList.push_back(myPCluster);
   }
   
   bool extended = false;
   bool isCrossing,isOverlapping;
   
   // try to cluster all new patterns in newClusterList
   for (ClusterListITER myNewCluster = newClusterList.begin();myNewCluster != newClusterList.end();++myNewCluster)
   {
      bool 				found = false;
      ClusterInfoPTR 	maxClus;

      // loop over all inside minBounds
      for (unsigned int actMinBound = 0;actMinBound < (*myNewCluster)->clusterMaxKey->minBounds.size();++actMinBound)
      {  
         int max 		= 0;
         int actmax 	= 0;
         int myDist	= INT_MAX;
             
         ClusterListITER before_begin 	= clusterSequence[(*myNewCluster)->clusterMaxKey->minBounds[actMinBound]-1]->clusterList.begin();
         ClusterListITER before_end 	= clusterSequence[(*myNewCluster)->clusterMaxKey->minBounds[actMinBound]-1]->clusterList.end();
						
         // loop over all before clusters i
         for (ClusterListITER i = before_begin;i != before_end;++i)
         {
            myDist = distance(*myNewCluster,*i,actMinBound,isCrossing,isOverlapping,true);
            
            int diff = abs(myDist - (*i)->actDistance);
            
            if ( (myDist < DISTANCE_THRESHOLD) && (diff <= DISTANCE_EQUAL_THRESHOLD) )
            {        
               actmax = (*i)->clusterSize;
               switch (CLUSTER_METHODE)
               {
                  case CLUSTER_ALL:
                     // (*myPos).clusterList.push_back(extCluster);
                     break;
		// cluster max 2 muss testen ob cluster schon geclustered wurde...
                  case CLUSTER_MAX_1: case CLUSTER_MAX_2:
                     if ((actmax>max))
                     {
                        maxClus = *i;
                        max = actmax;
                        found = true;
                     }
                     break;
               }
            }
         } //end loop Clusterlist pos before
         
         // extend actual inside bound with found cluster(s) within treshold   
         switch (found)
         {
         	case true:
         		switch (CLUSTER_METHODE)
         		{
         			case CLUSTER_MAX_2:
         				maxClus->clustered = true;
         			case CLUSTER_MAX_1:
         				if (maxClus->clusterMinMax.first < (*myNewCluster)->clusterMinMax.first)
         				{
         					(*myNewCluster)->clusterMinKey         = maxClus->clusterMinKey;
         					(*myNewCluster)->clusterMinMax.first   = maxClus->clusterMinMax.first;
         				}
         				(*myNewCluster)->clusterPatternList.insert((*myNewCluster)->clusterPatternList.end(),maxClus->clusterPatternList.begin(),maxClus->clusterPatternList.end());                    
         				(*myNewCluster)->clusterSize             += max;
         				(*myNewCluster)->actDistance             = 0;
         				(*myNewCluster)->lastClusterList.push_back(maxClus);
         				found = false;
         				extended = true;   
         			break;
         			case CLUSTER_ALL:;
         		} // switch CLUSTER_METHODE
         	case false:;
         } // end switch found
      } // end loop over all boundaries from new pattern
         
      if ((extended)) 
      {
         ++clusterCounter;
         (*myNewCluster)->clusterId  = "cl_"+ biu::Util_String::int2str(clusterCounter);
         extended = false;
         myPos->clusterList.push_back(*myNewCluster); 
      }
      else 
    	  delete *myNewCluster;
   } // end loop newClusterList
}

void mcsClustering::clusterToCluster(const ClusterNucleotidePTR& posStemEnd)
{
   bool					isCrossing,isOverlap;
   ClusterListTYPE 		newClusterList;
   int 					actBeforeClusterListPos;
   
   for(ClusterListITER actStemEndCluster = posStemEnd->clusterList.begin();actStemEndCluster != posStemEnd->clusterList.end();++actStemEndCluster)
   {
	   // test if cluster starts within stem or before stem
	   if ((*actStemEndCluster)->clusterMinMax.first > posStemEnd->connectedToNuc->beforeNuc->nucleotide->position)
		   actBeforeClusterListPos = posStemEnd->connectedToNuc->beforeNuc->nucleotide->position;
	   else
		   actBeforeClusterListPos = (*actStemEndCluster)->clusterMinMax.first - 1;
	   
	/*   if ((*actStemEndCluster)->clusterMinMax.first > posStemEnd->nucleotide->position)
	   		   actBeforeClusterListPos = posStemEnd->connectedToNuc->beforeNuc->nucleotide->position;
	   	   else
	   		   actBeforeClusterListPos = posStemEnd->nucleotide->position;*/
	   
	   ClusterInfoPTR maxClus	= NULL;
	   int max 					= 0;
	   int actMax 				= 0;
	   int myDist 				= INT_MAX;
	   bool found				= false;

	   for(ClusterListITER actBeforeCluster = clusterSequence[actBeforeClusterListPos]->clusterList.begin();actBeforeCluster != clusterSequence[actBeforeClusterListPos]->clusterList.end();++actBeforeCluster)
	   {
		   myDist = distance(*actStemEndCluster,*actBeforeCluster,-1,isCrossing,isOverlap,false);
		   
		   int diff = abs(myDist - (((int)(*actStemEndCluster)->clusterMinMax.first) - ((int)(*actBeforeCluster)->clusterMinMax.second)));
		   
		   //if ((myDist < DISTANCE_TRESHOLD)&&((*actBeforeCluster)->actDistance + (*actStemEndCluster)->actDistance <= DISTANCE_TRESHOLD))
		   if ( (myDist < DISTANCE_THRESHOLD) && (( (int)((*actStemEndCluster)->clusterMinMax.first) - (int)((*actBeforeCluster)->clusterMinMax.second)) <= DISTANCE_THRESHOLD) && (diff <= DISTANCE_EQUAL_THRESHOLD) )  		   
		   {
			   actMax = (*actBeforeCluster)->clusterSize;
			   switch (CLUSTER_METHODE)
			   {
			   	case CLUSTER_ALL:
			   		break;
			   	case CLUSTER_MAX_1: case CLUSTER_MAX_2:
			   		if (actMax>max)
			   		{
			   			maxClus = *actBeforeCluster;
			   			max = actMax;
			   			found = true;
			   		}
			   		break;
			   } // end switch CLUSTER_METHODE
		   } // end if distance test    
	   } // end loop clusters from posBeforeStem
	   
	   switch (CLUSTER_METHODE)
	   {
	   	case CLUSTER_MAX_2:
	   		if (found) maxClus->clustered = true;
	   	case CLUSTER_MAX_1:
	   		if (found)
	   		{
	   			++clusterCounter;
	   			ClusterInfoPTR extCluster        	= ClusterInfoPTR(new ClusterInfoTYPE);
	   			*extCluster                      	= *maxClus;
	   			extCluster->clusterId            	= "cl_"+ biu::Util_String::int2str(clusterCounter);          
	   			extCluster->clusterMaxKey				= (*actStemEndCluster)->clusterMaxKey;           
	   			extCluster->clusterMinMax.second  	= (*actStemEndCluster)->clusterMinMax.second;        
	   			extCluster->clusterPatternList.insert(extCluster->clusterPatternList.end(),(*actStemEndCluster)->clusterPatternList.begin(),(*actStemEndCluster)->clusterPatternList.end());
	   			extCluster->clusterSize          	+= (*actStemEndCluster)->clusterSize;
	   			//beachten!!! shortest dist oder seq dist hier unterscheiden ??? nein,da dist zu max pattern entscheidend
	   			//extCluster->actDistance          	= min((*posBeforeCluster)->actDistance+2,maxClus->actDistance);
	   			extCluster->actDistance          	= (*actStemEndCluster)->actDistance;
	   			extCluster->lastClusterList.push_back(*actStemEndCluster);
	   			newClusterList.push_back(extCluster);
	   			found = false;
	   		}
	   		break;
	   	case CLUSTER_ALL:;
	   } // end switch CLSUTER_METHODE
   }// end for loop actStemEndCluster

   (*posStemEnd).clusterList.insert((*posStemEnd).clusterList.end(),newClusterList.begin(),newClusterList.end());
}

// test if cluster1 after/outside cluster2 ???, in first structure cluster1 is after/outside cluster2
int mcsClustering::distance(  const ClusterInfoPTR& cluster1,
                              const ClusterInfoPTR& cluster2,
                              const int testBound,
                              bool& crossing, bool& overlapping, bool isPosBeforeTest)
{
   int dist = INT_MAX;
  // int tmpDist1,tmpDist2,min,max;
   crossing = false;
   overlapping = false;
   bool outsideStr1 = false;
   bool outsideStr2 = false;
   bool insideStr1 = false;
   bool insideStr2 = false;
           	   
   // crossing/overlapping test, if ok then calculate distance in second structure
   switch (isPosBeforeTest)
   {
    // test from clusterActPatterns,   
   	case true:
    	  
         switch (cluster1->clusterMaxKey->patternType)
         {
           case HAIRPINPATTERN: 
        	   crossing = true; 
        	   overlapping = false;
        	   return INT_MAX; 
        	   break;
           case LOOPPATTERN:
        	   // closing bond beachten!!!
        	   
        	   if ( (cluster1->clusterMinKey->lastClosingBond.first <= cluster2->clusterMaxKey->lastClosingBond.first) &&
        	        (cluster1->clusterMaxKey->lastClosingBond.second >= cluster2->clusterMaxKey->lastClosingBond.second) )
        		 outsideStr1 = true;
        	   else outsideStr1 = false;
        	   if ( (cluster1->clusterMinKey->lastClosingBondREF.first <= cluster2->clusterMaxKey->lastClosingBondREF.first) &&
           	        (cluster1->clusterMaxKey->lastClosingBondREF.second >= cluster2->clusterMaxKey->lastClosingBondREF.second) )
        		 outsideStr2 = true;
        	   else outsideStr2 = false;
        	   
        	   if ( (cluster1->clusterMinKey->lastClosingBond.first > cluster2->clusterMaxKey->lastClosingBond.first) &&
        	           	        (cluster1->clusterMaxKey->lastClosingBond.second < cluster2->clusterMaxKey->lastClosingBond.second) )
        	           		 insideStr1 = true;
        	           	   else insideStr1 = false;
        	           	   if ( (cluster1->clusterMinKey->lastClosingBondREF.first > cluster2->clusterMaxKey->lastClosingBondREF.first) &&
        	              	        (cluster1->clusterMaxKey->lastClosingBondREF.second < cluster2->clusterMaxKey->lastClosingBondREF.second) )
        	           		 insideStr2 = true;
        	           	   else insideStr2 = false;
        	   
        	   if ( ( ((outsideStr1) && (outsideStr2)) || (!outsideStr1) && (!outsideStr2) ) || 
        	        ( ((insideStr1) && (insideStr2)) || (!insideStr1) && (!insideStr2) ) )
        		crossing = false;
        	   else crossing = true;
        		   
               if ( (cluster1->clusterMinKey->minmaxREF.first <= cluster2->clusterMaxKey->minmaxREF.second) || (crossing))
               {
            	  crossing = true; 
                  overlapping = false;
                  return INT_MAX;
               }
               break;
            case STEMLOOPPATTERN:
               if (!((cluster2->clusterMinKey->minmax.first  > cluster1->clusterMaxKey->maxBounds[testBound])&&
                      (cluster2->clusterMinKey->minmax.second < cluster1->clusterMaxKey->minBounds[testBound])))                       
               {
                  crossing = true; 
                  overlapping = false;
                  return INT_MAX;
               }
               //Achtung! Hier kein break, test geht nach unten weiter!
            case STEMPATTERN:
               if (!((cluster1->clusterMinKey->maxBoundsREF[testBound] < cluster2->clusterMinKey->minmaxREF.first)&&
                     (cluster1->clusterMaxKey->minBoundsREF[testBound] > cluster2->clusterMaxKey->minmaxREF.second)))
               {
                  crossing = true; 
                  overlapping = false;
                  return INT_MAX;
               } 
               break;
         }
         // distance
         //min = cluster2->clusterMaxKey->minmaxREF.first;
         //max = cluster2->clusterMaxKey->minmaxREF.second; 

        // tmpDist1 = abs((int(cluster1->clusterMinKey->maxBoundsREF[testBound]))-(int(min)));
         dist = abs((int(cluster1->clusterMinKey->minBoundsREF[testBound]))-(int(cluster2->clusterMaxKey->minmaxREF.second)));
               		
         //if (tmpDist1 < tmpDist2) dist = tmpDist1;
         	/*else*/ //dist = tmpDist2;
         break;
      
      case false:
    	  if ( (cluster1->clusterMinKey->lastClosingBond.first <= cluster2->clusterMaxKey->lastClosingBond.first) &&
    	          	        (cluster1->clusterMaxKey->lastClosingBond.second >= cluster2->clusterMaxKey->lastClosingBond.second) )
    	          		 outsideStr1 = true;
    	          	   else outsideStr1 = false;
    	  
    	  			//if cluster1->clusterMinKey
    	  
    	          	   if ( (cluster1->clusterMinKey->lastClosingBondREF.first <= cluster2->clusterMaxKey->lastClosingBondREF.first) &&
    	             	        (cluster1->clusterMaxKey->lastClosingBondREF.second >= cluster2->clusterMaxKey->lastClosingBondREF.second) )
    	          		 outsideStr2 = true;
    	          	   else outsideStr2 = false;
    	          	   
    	          	 if ( (cluster1->clusterMinKey->lastClosingBond.first > cluster2->clusterMaxKey->lastClosingBond.first) &&
    	          	         	           	        (cluster1->clusterMaxKey->lastClosingBond.second < cluster2->clusterMaxKey->lastClosingBond.second) )
    	          	         	           		 insideStr1 = true;
    	          	         	           	   else insideStr1 = false;
    	          	         	           	   if ( (cluster1->clusterMinKey->lastClosingBondREF.first > cluster2->clusterMaxKey->lastClosingBondREF.first) &&
    	          	         	              	        (cluster1->clusterMaxKey->lastClosingBondREF.second < cluster2->clusterMaxKey->lastClosingBondREF.second) )
    	          	         	           		 insideStr2 = true;
    	          	         	           	   else insideStr2 = false;
    	          	         	   
    	          	 if ( ( ((outsideStr1) && (outsideStr2)) || (!outsideStr1) && (!outsideStr2) ) || 
    	          	      ( ((insideStr1) && (insideStr2)) || (!insideStr1) && (!insideStr2) ) )
    	          		 crossing = false;
    	          	 else crossing = true;
    	          	 
    	        
    	          	 
    	  		if ( (!(cluster1->clusterMinKey->minmaxREF.first > cluster2->clusterMaxKey->minmaxREF.second)) || (crossing)/*&&
    	  		    (cluster1->clusterMinKey->minmaxREF.second > cluster2->clusterMaxKey->minmaxREF.second)*/)
    	  		{
    	  			crossing = true; 
    	  		    overlapping = false;
    	  		    return INT_MAX;
    	  		}
    	  		         
    	  		// f�r looppattern closing bond beachten...
    	  		         
    	  		// distance
    	  		//max = cluster2->clusterMaxKey->minmaxREF.second;
    	  		dist = abs((int(cluster1->clusterMinKey->minmaxREF.first))-(int(cluster2->clusterMaxKey->minmaxREF.second)));
   }
   return dist;
}

void mcsClustering::traverseCluster(ClusterInfoPTR myClus)
{
   myClus->visited = true;
   if (myClus->clusterPatternList.size()>1)
   {
      for (ClusterListCITER nextCluster = myClus->lastClusterList.begin();nextCluster != myClus->lastClusterList.end();++nextCluster) 
      {  
         traverseCluster(*nextCluster);
      }
   }
}

void mcsClustering::collectClusters()
{
   for (ClusterSequenceTYPE::reverse_iterator myPos=clusterSequence.rbegin(); myPos != clusterSequence.rend();++myPos)    
   {
      for(ClusterListCITER myCluster = (*myPos)->clusterList.begin();myCluster != (*myPos)->clusterList.end();++myCluster)
      {
         if ((!(*myCluster)->visited) && ((*myCluster)->clusterPatternList.size()>1))
         {
            clusters.insert(make_pair((*myCluster)->clusterSize,*myCluster));
            traverseCluster(*myCluster);
         }
      }
   }
   cout << "Collected Clusters: " << clusters.size() << endl;
}

// -------------------------------------------------------------------------------------------------------
//
//	following functions only for info and debugging !!!
//
// -------------------------------------------------------------------------------------------------------
void mcsClustering::printClusterInfo()
{
   cout << "ClusterCouter: " << clusterCounter << endl;
   
  /* for (ClusterStrandTYPE::reverse_iterator myPos=clusterStrand.rbegin(); myPos != clusterStrand.rend();++myPos)    
   {
      cout << "---------------------------------------------------" << endl;
      cout << (*myPos).nucleotide->position << " ClusterStrand Position " << endl;
      cout << "BaseType: " << (*myPos).baseType << endl;
      cout << "ClusterListSize: " << (*myPos).clusterList.size() << endl;
  */ /* lassen  cout << "PatternKeySize: "  << mol1.clusterKeyStrand[(*myPos).nucleotide->position].maxKeys.size() << "\tNew Pats: ";
      for (PosKeyListCITER myPat = mol1.clusterKeyStrand[(*myPos).nucleotide->position].maxKeys.begin();myPat != mol1.clusterKeyStrand[(*myPos).nucleotide->position].maxKeys.end();++myPat)
      {
         cout << (*myPat)->patternId << "\t";
      } lassen */
  /*    cout << endl;
      cout << "ClosingBond: " << (*myPos).nucleotide->lastClosingBond.first <<","<< (*myPos).nucleotide->lastClosingBond.second<<endl;
      for (ClusterListCITER myCluster= (*myPos).clusterList.begin();myCluster != (*myPos).clusterList.end();++myCluster)
      {
         cout << "\t size:" << (*myCluster)->clusterSize << "\tid: " << (*myCluster)->clusterId << "\tPatterns: "<<(*myCluster)->clusterPatternList.size()
              <<"\tPatId: "<<(*myCluster)->newPattern->patternId<< "\tmin,max: "<< (*myCluster)->clusterMinMax.first<<","<<(*myCluster)->clusterMinMax.second
              <<"\tPatMinMaREFx: "<<(*myCluster)->newPattern->minmaxREF.first<<","<<(*myCluster)->newPattern->minmaxREF.second<<"\tdist: " << (*myCluster)->actDistance 
              << endl;
          cout << "\t\t\tPatIds: ";
         for (PosKeyListCITER myPat = (*myCluster)->clusterPatternList.begin();myPat != (*myCluster)->clusterPatternList.end();++myPat)
         {
            cout << (*myPat)->patternId << "\t"; 
         }
         cout << endl <<endl;
      }      
      cout << endl;
   }*/
}

/*void mcsClustering::printClusterList(const BaseClusterListTYPE myList,bool printListDetails)
{
   cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> CLUSTERLIST >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"<<endl;
   ClusterMolecule1::printNucleotide(myList.nucleotide);
   cout << "\tBaseType: "<< myList.baseType<<endl;
  // cout << "\tMaxKeyListSize: " << mol1.clusterKeyStrand[myList.nucleotide->position].maxKeys.size()<<endl;
   cout << "\tList Size: " << myList.clusterList.size() << "\tList contains ClusterIds: ";
   for (ClusterListCITER myC=myList.clusterList.begin();myC!=myList.clusterList.end();++myC)
   {
      cout << (*myC)->clusterId <<"\t";
   }
   if (printListDetails) 
   for (ClusterListCITER myC=myList.clusterList.begin();myC!=myList.clusterList.end();++myC)
   {
      printCluster(*myC);
   }
   cout << "\n--------------------------------------------------------------------"<<endl;
}*/

void mcsClustering::printCluster(const ClusterInfoPTR myClus)
{
   ClusterInfoTYPE myC = *myClus;
   cout <<"\n\t\t--------------------------------------------------" << endl;
   cout << "\t\t>>>>>>> Cluster Info <<<<<<<<<<<<<<<<<<<<<<<<<<<<< " << endl;
   cout <<"\t\t--------------------------------------------------" << endl;
   cout <<"\t\t\tClusterId   :\t"<< myC.clusterId <<endl;
   cout <<"\t\t\t\tClusterSize :" << myC.clusterSize << "\tid: " << myC.clusterId << "\tPatterns: "<< myC.clusterPatternList.size()
              <<"\tPatId: "<<myC.newPattern->patternId<< "\tmin,max: "<< myC.clusterMinMax.first<<","<<myC.clusterMinMax.second
              <<"\tPatMinMax: "<<myC.newPattern->minmax.first<<","<<myC.newPattern->minmax.second
              <<"\tPatMinMaxREF: "<<myC.newPattern->minmaxREF.first<<","<<myC.newPattern->minmaxREF.second
              <<"\tPatMinKey: "<<myC.clusterMinKey->patternId<<"\tPatMaxKey: "<<myC.clusterMaxKey->patternId
              <<"\tdist: " << myC.actDistance 
              << endl;
   cout << "\t\t\t\t\tPatIds: ";
   for (PatternKeyListCITER myPat = myC.clusterPatternList.begin();myPat != myC.clusterPatternList.end();++myPat)
   {
      cout << (*myPat)->patternId << "\t"; 
   }
   cout<<endl;
   //cout << "\n\t\t<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n"   ;
}

void mcsClustering::printPatInfo(const SinglePattern& myPat1,const SinglePattern& myPat2)
{
   cout << "\n-->> Pattern: " << myPat1.getID().getId() << endl;
   intVec pat1,pat2;
   pat1 = myPat1.getPat();
   pat2 = myPat2.getPat();
    cout << "  pos: " << setw(4) << right << pat1[0]-1 << " " << mol1.getBase(pat1[0]-1) << " " << mol1.getStructure(pat1[0]-1) << " " << mol1.getStructure(mol1.getBasePartner(pat1[0]-1)) << " " << mol1.getBase(mol1.getBasePartner(pat1[0]-1)) <<" " <<mol1.getBasePartner(pat1[0]-1)
              << "\t\t" << setw(4) << right << pat2[0]-1 << " " << mol2.getBase(pat2[0]-1) << " " << mol2.getStructure(pat2[0]-1) << " " << mol2.getStructure(mol2.getBasePartner(pat2[0]-1)) << " " << mol2.getBase(mol2.getBasePartner(pat2[0]-1)) << endl;
              
   cout <<"  ------- pattern ---------"<<endl;
   for (unsigned int i=0;i<pat1.size();i++)
      {
         cout << "  pos: " << setw(4) << right << pat1[i] << " " << mol1.getBase(pat1[i]) << " " << mol1.getStructure(pat1[i]) << " " << mol1.getStructure(mol1.getBasePartner(pat1[i])) << " " << mol1.getBase(mol1.getBasePartner(pat1[i])) <<" " <<mol1.getBasePartner(pat1[i])
              << "\t\t" << setw(4) << right << pat2[i] << " " << mol2.getBase(pat2[i]) << " " << mol2.getStructure(pat2[i]) << " " << mol2.getStructure(mol2.getBasePartner(pat2[i])) << " " << mol2.getBase(mol2.getBasePartner(pat2[i])) << endl;
      }
  cout <<"  ------- pattern ---------"<<endl;    
  cout << "  pos: " << setw(4) << right << pat1[0]+1 << " " << mol1.getBase(pat1[0]+1) << " " << mol1.getStructure(pat1[0]+1) << " " << mol1.getStructure(mol1.getBasePartner(pat1[0]+1)) << " " << mol1.getBase(mol1.getBasePartner(pat1[0]+1)) <<" " <<mol1.getBasePartner(pat1[0]+1)
              << "\t\t" << setw(4) << right << pat2[pat2.size()-1]+1 << " " << mol2.getBase(pat2[pat2.size()-1]+1) << " " << mol2.getStructure(pat2[pat2.size()-1]+1) << " " << mol2.getStructure(mol2.getBasePartner(pat2[pat2.size()-1]+1)) << " " << mol2.getBase(mol2.getBasePartner(pat2[pat2.size()-1]+1)) << endl;
    
}
                  
// namespace clustering
